use nix::poll::{poll, PollFd, PollFlags};
use std::{cmp, error::Error as StdError, fmt, time::{Duration, Instant}, os::unix::prelude::AsRawFd};
use xcb::{x, Xid};

use bscreensaver_util::{create_atom, BSCREENSAVER_WM_CLASS};

const COMMAND_WINDOW_ID_ATOM_NAME: &[u8] = b"BSCREENSAVER_COMMAND_WINDOW_ID";
const COMMAND_WINDOW_WM_NAME: &[u8] = b"bscreensaver command window";

const BSCREENSAVER_BLANK_ATOM_NAME: &[u8] = b"BSCREENSAVER_BLANK";
const BSCREENSAVER_LOCK_ATOM_NAME: &[u8] = b"BSCREENSAVER_LOCK";
const BSCREENSAVER_DEACTIVATE_ATOM_NAME: &[u8] = b"BSCREENSAVER_DEACTIVATE";
const BSCREENSAVER_RESTART_ATOM_NAME: &[u8] = b"BSCREENSAVER_RESTART";
const BSCREENSAVER_EXIT_ATOM_NAME: &[u8] = b"BSCREENSAVER_EXIT";

const BSCREENSAVER_RESULT_ATOM_NANE: &[u8] = b"BSCREENSAVER_RESULT";

#[derive(Debug, Clone, Copy)]
pub enum BCommand {
    Blank,
    Lock,
    Deactivate,
    Restart,
    Exit,
}

impl BCommand {
    pub fn atom_name(&self) -> &'static [u8] {
        match self {
            Self::Blank => BSCREENSAVER_BLANK_ATOM_NAME,
            Self::Lock => BSCREENSAVER_LOCK_ATOM_NAME,
            Self::Deactivate => BSCREENSAVER_DEACTIVATE_ATOM_NAME,
            Self::Restart => BSCREENSAVER_RESTART_ATOM_NAME,
            Self::Exit => BSCREENSAVER_EXIT_ATOM_NAME,
        }
    }
}

#[derive(Debug)]
pub enum Error {
    X(xcb::Error),
    Sys(nix::errno::Errno),
    NotRunning,
    CommandError,
    Timeout,
}

impl fmt::Display for Error {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        match self {
            Self::X(err) => write!(f, "X error: {}", err),
            Self::Sys(err) => write!(f, "System error: {}", err),
            Self::NotRunning => write!(f, "bscreensaver is not running"),
            Self::CommandError => write!(f, "bscreensaver returned an error"),
            Self::Timeout => write!(f, "Timed out waiting for a response"),
        }
    }
}

impl StdError for Error {
    fn source(&self) -> Option<&(dyn StdError + 'static)> {
        match self {
            Self::X(err) => Some(err),
            Self::Sys(err) => Some(err),
            Self::NotRunning => None,
            Self::CommandError => None,
            Self::Timeout => None,
        }
    }
}

impl From<xcb::Error> for Error {
    fn from(error: xcb::Error) -> Self {
        Self::X(error)
    }
}

impl From<xcb::ProtocolError> for Error {
    fn from(error: xcb::ProtocolError) -> Self {
        Self::X(xcb::Error::Protocol(error))
    }
}

impl From<xcb::ConnError> for Error {
    fn from(error: xcb::ConnError) -> Self {
        Self::X(xcb::Error::Connection(error))
    }
}

impl From<nix::errno::Errno> for Error {
    fn from(error: nix::errno::Errno) -> Self {
        Self::Sys(error)
    }
}

pub fn create_command_window(conn: &xcb::Connection, screen: &x::Screen) -> Result<x::Window, Error> {
    let mut cookies = Vec::new();

    let msg_win: x::Window = conn.generate_id();
    cookies.push(conn.send_request_checked(&x::CreateWindow {
        depth: x::COPY_FROM_PARENT as u8,
        wid: msg_win,
        parent: screen.root(),
        x: -50,
        y: -50,
        width: 1,
        height: 1,
        border_width: 0,
        class: x::WindowClass::InputOnly,
        visual: x::COPY_FROM_PARENT,
        value_list: &[
               x::Cw::OverrideRedirect(true),
               x::Cw::EventMask(x::EventMask::STRUCTURE_NOTIFY | x::EventMask::PROPERTY_CHANGE),
        ],
    }));
    cookies.push(conn.send_request_checked(&x::ChangeProperty {
        mode: x::PropMode::Replace,
        window: msg_win,
        property: x::ATOM_WM_NAME,
        r#type: x::ATOM_STRING,
        data: COMMAND_WINDOW_WM_NAME,
    }));
    cookies.push(conn.send_request_checked(&x::ChangeProperty {
        mode: x::PropMode::Replace,
        window: msg_win,
        property: x::ATOM_WM_CLASS,
        r#type: x::ATOM_STRING,
        data: BSCREENSAVER_WM_CLASS,
    }));

    let msg_win_atom = create_atom(&conn, COMMAND_WINDOW_ID_ATOM_NAME)?;
    cookies.push(conn.send_request_checked(&x::ChangeProperty {
        mode: x::PropMode::Replace,
        window: screen.root(),
        property: msg_win_atom,
        r#type: x::ATOM_WINDOW,
        data: &[msg_win.resource_id()],
    }));

    for cookie in cookies {
        conn.check_request(cookie)?;
    }

    Ok(msg_win)
}

pub fn bscreensaver_command_response(conn: &xcb::Connection, reply_window: x::Window, is_success: bool) -> Result<(), Error> {
    let message_type = create_atom(conn, BSCREENSAVER_RESULT_ATOM_NANE)?;
    conn.send_and_check_request(&x::SendEvent {
        propagate: false,
        destination: x::SendEventDest::Window(reply_window),
        event_mask: x::EventMask::STRUCTURE_NOTIFY,
        event: &x::ClientMessageEvent::new(
            reply_window,
            message_type,
            x::ClientMessageData::Data32([
                if is_success { 1 } else { 0 },
                0,
                0,
                0,
                0,
            ])
        ),
    })?;
    Ok(())
}

pub fn bscreensaver_command(command: BCommand, timeout: Option<Duration>) -> Result<(), Error> {
    let (conn, _) = xcb::Connection::connect(None)?;
    let setup = conn.get_setup();

    let msg_window_id_atom = create_atom(&conn, COMMAND_WINDOW_ID_ATOM_NAME)?;
    let (msg_window_screen, msg_window_id) = 'outer: loop {
        for screen in setup.roots() {
            let cookie = conn.send_request(&x::GetProperty {
                delete: false,
                window: screen.root(),
                property: msg_window_id_atom,
                r#type: x::ATOM_WINDOW,
                long_offset: 0,
                long_length: std::mem::size_of::<u32>() as u32,
            });
            let reply = conn.wait_for_reply(cookie)?;
            let windows = reply.value::<x::Window>();
            if windows.is_empty() {
                continue;
            }
            let window = windows[0];

            let cookie = conn.send_request(&x::GetProperty {
                delete: false,
                window,
                property: x::ATOM_WM_NAME,
                r#type: x::ATOM_STRING,
                long_offset: 0,
                long_length: COMMAND_WINDOW_WM_NAME.len() as u32 + 1,
            });
            let reply = conn.wait_for_reply(cookie)?;
            if reply.value::<u8>() == COMMAND_WINDOW_WM_NAME {
                break 'outer Ok((screen, window));
            }
        }
        break Err(Error::NotRunning);
    }?;
    let command_atom = create_atom(&conn, command.atom_name())?;
    let result_atom = create_atom(&conn, BSCREENSAVER_RESULT_ATOM_NANE)?;

    let reply_window = conn.generate_id();
    conn.send_and_check_request(&x::CreateWindow {
        depth: x::COPY_FROM_PARENT as u8,
        wid: reply_window,
        parent: msg_window_screen.root(),
        x: -50,
        y: -50,
        width: 1,
        height: 1,
        border_width: 0,
        class: x::WindowClass::InputOnly,
        visual: x::COPY_FROM_PARENT,
        value_list: &[
               x::Cw::OverrideRedirect(true),
               x::Cw::EventMask(x::EventMask::STRUCTURE_NOTIFY),
        ],
    })?;

    let res = conn.send_and_check_request(&x::SendEvent {
        propagate: false,
        destination: x::SendEventDest::Window(msg_window_id),
        event_mask: x::EventMask::STRUCTURE_NOTIFY,
        event: &x::ClientMessageEvent::new(
            msg_window_id,
            command_atom,
            x::ClientMessageData::Data32([
                reply_window.resource_id(),
                0,
                0,
                0,
                0,
            ])
        ),
    });

    match res {
        Err(xcb::ProtocolError::X(x::Error::Window(x::WindowError { .. }), _)) => Err(Error::NotRunning)?,
        Err(err) => Err(Error::X(xcb::Error::Protocol(err)))?,
        Ok(_) => (),
    }

    let start = Instant::now();
    'outer1: loop {
        while let Some(event) = conn.poll_for_event()? {
            match event {
                xcb::Event::X(x::Event::ClientMessage(ev)) if ev.r#type() == result_atom => {
                    match ev.data() {
                        x::ClientMessageData::Data32(data) => {
                            if data[0] != 1 {
                                break 'outer1 Err(Error::CommandError);
                            }
                        },
                        _ => break 'outer1 Err(Error::CommandError),
                    }
                    break 'outer1 Ok(());
                },
                _ => (),
            }
        }

        let poll_timeout = timeout.map(|to| {
            let since_start = start.elapsed();
            if since_start > to {
                0i32
            } else {
                cmp::max(i32::MAX as u128, (to - since_start).as_millis()) as i32
            }
        }).unwrap_or(-1);
        if poll_timeout == 0 {
            break 'outer1 Err(Error::Timeout);
        }

        let mut pfds = vec![PollFd::new(conn.as_raw_fd(), PollFlags::POLLIN)];
        poll(pfds.as_mut_slice(), poll_timeout)?;
    }
}
